from collections.abc import Mapping
import json
from typing import Any, TypeVar
import warnings

from django import template

register = template.Library()


_KT = TypeVar("_KT")
_VT_co = TypeVar("_VT_co", covariant=True)


@register.filter
def get_item(value: Mapping[_KT, _VT_co], arg: _KT) -> _VT_co | None:
    """
    Retrieve `value[arg]` from a mapping `value`, where `arg` can be a variable.

    This will return `None` if `arg` is not found. It may be useful to chain the `default_if_none`
    filter after this one.

    Sample usage::
        {% load resonant_utils %}
        {% for key in some_keys %}
            {{ my_dict|getitem:key }}
        {% endfor %}
    """
    return value.get(arg, None)


@register.filter
def getitem(value: Mapping[_KT, _VT_co], arg: _KT) -> _VT_co | None:
    warnings.warn(
        'deprecated: use "get_item" instead of "getitem".', DeprecationWarning, stacklevel=2
    )
    return get_item(value, arg)


@register.filter
def pretty_json(value: Any, indent: int | None = None) -> str:
    """
    Convert `value` to a JSON-formatted string.

    Optionally, `indent` can be specified as an positive integer number of spaces to pretty-print
    indentation with; `None` (the default) will disable pretty-printing.

    The output should typically be embedded within HTML an `<pre>` element.
    If `indent` is specified, the output will likely contain newlines, which `<pre>` will render.

    Sample usage::
    {% load resonant_utils %}
    <pre>{{ my_object|pretty_json:4 }}</pre>
    """
    return json.dumps(value, indent=indent)



from collections.abc import Sequence
import json
from typing import Any, TypeVar

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Manager, Model, QuerySet
from django.utils.translation import gettext_lazy as _

_T = TypeVar("_T", bound=Model, covariant=True)


class JSONObjectField(models.JSONField):
    """
    A field for storing JSON Objects.

    Other JSON types are not allowed at the top level.
    """

    @staticmethod
    def _validate_is_object(val: Any) -> None:
        """Validate that the value is a JSON Object (dict) at the top level."""
        if not isinstance(val, dict):
            raise ValidationError(_("Must be a JSON Object."))

    empty_values: list[Any] = [{}]

    # At this point in the class definition lifecycle,
    # staticmethods aren't resolvable directly, so use "__func__"
    # https://stackoverflow.com/questions/41921255/staticmethod-object-is-not-callable
    # Additionally, MyPy has a bug here: https://github.com/python/mypy/issues/3482
    default_validators = [_validate_is_object.__func__]  # type: ignore[attr-defined]

    def __init__(
        self,
        verbose_name: str | None = None,
        name: str | None = None,
        encoder: type[json.JSONEncoder] | None = None,
        decoder: type[json.JSONDecoder] | None = None,
        **kwargs: Any,
    ) -> None:
        kwargs["default"] = dict
        kwargs["blank"] = True
        super().__init__(
            verbose_name=verbose_name, name=name, encoder=encoder, decoder=decoder, **kwargs
        )

    def deconstruct(self) -> tuple[str, str, Sequence[Any], dict[str, Any]]:
        name, path, args, kwargs = super().deconstruct()
        del kwargs["default"]
        del kwargs["blank"]
        return name, path, args, kwargs


class SelectRelatedManager(Manager[_T]):
    """
    A Manager which always follows specified foreign-key relationships.

    For Models which very frequently need to follow foreign-key relationships
    (e.g. to generate their own __str__), it may be best to make this the
    "_default_manager", so many automatically generated queries are more
    efficient. To do so, define a custom "objects" as the first
    Manager in the Model:
        objects = SelectRelatedManager('foreign_thing')

    Passing no arguments makes this follow all the Model's non-null
    foreign-key relationships.
    """

    def __init__(self, *related_fields: str) -> None:
        self.related_fields: list[str] = list(related_fields)
        super().__init__()

    def get_queryset(self) -> QuerySet[_T]:
        return super().get_queryset().select_related(*self.related_fields)


class DeferredFieldsManager(Manager[_T]):
    """A Manager which defers loading specified fields within fetched Models."""

    def __init__(self, *deferred_fields: str) -> None:
        self.deferred_fields = deferred_fields
        super().__init__()

    def get_queryset(self) -> QuerySet[_T]:
        return super().get_queryset().defer(*self.deferred_fields)



from collections.abc import Callable
import itertools
from typing import TYPE_CHECKING, Generic, TypeVar

from django.contrib import admin
from django.contrib.admin.options import InlineModelAdmin
from django.db.models import Model
from django.http import HttpRequest

_ChildModelT = TypeVar("_ChildModelT", bound=Model)
_ParentModelT = TypeVar("_ParentModelT", bound=Model)

# https://mypy.readthedocs.io/en/latest/runtime_troubles.html#using-classes-that-are-generic-in-stubs-but-not-at-runtime
if TYPE_CHECKING:

    class _InlineMixin(InlineModelAdmin[_ChildModelT, _ParentModelT]):
        pass

    class _TabularInline(admin.TabularInline[_ChildModelT, _ParentModelT]):
        pass

else:

    class _InlineMixin(Generic[_ChildModelT, _ParentModelT]):
        pass

    class _TabularInline(Generic[_ChildModelT, _ParentModelT], admin.TabularInline):
        pass


class ReadonlyInlineMixin(_InlineMixin[_ChildModelT, _ParentModelT]):
    can_delete = False
    show_change_link = True
    view_on_site: bool | Callable[[_ChildModelT], str] = False
    extra = 0

    def get_readonly_fields(
        self, request: HttpRequest, obj: _ChildModelT | None = None
    ) -> list[str]:
        if self.fields is None:
            return []
        # Make all fields readonly
        # self.fields can contain nested sequences, but get_readonly_fields must return flattened
        return list(
            itertools.chain.from_iterable(
                [field_group] if isinstance(field_group, str) else field_group
                for field_group in self.fields
            )
        )

    def has_add_permission(self, request: HttpRequest, obj: _ParentModelT | None = None) -> bool:
        return False


class ReadonlyTabularInline(
    ReadonlyInlineMixin[_ChildModelT, _ParentModelT],
    _TabularInline[_ChildModelT, _ParentModelT],
):
    pass



from django import forms
from django.contrib.auth.models import AbstractUser
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _


# Django-Allauth further inherits from this form, and adds the typical "username", "email", etc.
# fields.
class FullNameSignupForm(forms.Form):
    """
    A Django-Allauth signup form which prompts for a user's first and last name.

    To use this, set the setting:
    ACCOUNT_SIGNUP_FORM_CLASS = 'resonant_utils.allauth.FullNameSignupForm'
    """

    first_name = forms.CharField(
        max_length=150,
        label=_("First name"),
        widget=forms.TextInput(
            attrs={"placeholder": _("First name"), "autocomplete": "given-name"}
        ),
    )
    last_name = forms.CharField(
        max_length=150,
        label=_("Last name"),
        widget=forms.TextInput(
            attrs={"placeholder": _("Last name"), "auto complete": "family-name"}
        ),
    )

    field_order = [
        # Any undefined fields in this list are just ignored
        "first_name",
        "last_name",
        "email",
        "email2",
        "username",
        "password1",
        "password2",
    ]

    def signup(self, request: HttpRequest, user: AbstractUser) -> None:
        # Allauth requires this method to be defined
        pass



import os
import sys

from django.core.management import ManagementUtility
from django.core.management.color import color_style


def execute_from_command_line(argv: list[str] | None = None) -> None:
    """Behave like `django.core.management.execute_from_command_line`, but print settings errors."""
    utility = ManagementUtility(argv)
    try:
        utility.execute()
    finally:
        if utility.settings_exception is not None and "DJANGO_AUTO_COMPLETE" not in os.environ:
            # A settings exception may exist after least 4 different code paths:
            # * Clean return from successfully printing help, which will have already printed the
            #   settings error, but we'll accept a second print here, for a simpler implementation.
            # * Clean return from running any other command
            # * `sys.exit(1)` if a command cannot be found (possibly because it's from an app,
            #   which aren't loaded if settings fail)
            # * `sys.exit(0)`, if autocomplete was invoked; we won't print here, to avoid spam (but
            #   we should still detect autocomplete natively, not hardcoding around `sys.exit(0)`)
            style = color_style()
            sys.stderr.write(
                style.NOTICE(
                    "Note that settings are not properly configured "
                    f"(error: {utility.settings_exception}).\n"
                )
            )



from __future__ import annotations

from django.http.response import HttpResponseRedirectBase
from django.middleware.common import CommonMiddleware


class HttpResponsePermanentRedirect(HttpResponseRedirectBase):
    status_code = 308


class PermanentRedirectCommonMiddleware(CommonMiddleware):
    response_redirect_class = HttpResponsePermanentRedirect
